from cStringIO import StringIO
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.DirectoryView import addDirectoryViews
from Products.PortalTransitions import ANY_TYPE
from Products.PortalTransitions.Extensions.WorkflowToolPatches import PatchedWorkflowTool
from Products.PortalTransitions.config import WorkflowToolClass, INSTALL_EXAMPLE, GLOBALS

def installTransitions(self, out):
    # install a few sample transitions
    tt = getToolByName(self, 'portal_transitions')
    info = {'type':ANY_TYPE, 'workflow':'plone_workflow', 'execute_after':1 }
    if not 'notify_submission' in tt.objectIds():
        info['id'] = 'notify_submission'
        info['transition'] = 'submit'
        # send an email to all reviewers when an item is submitted
        info['expression'] = 'python:portal.portal_transitions.backgroundSendMail(object, \'notify_submission\', roles=[\'Reviewer\'])'
        # alternatively, send an email to the admin
        # info['expression'] = 'python:portal.portal_transitions.backgroundSendMail(object, \'notify_submission\', ids=[\'admin\'])' }
        tt.manage_addExpression(info)
    if not 'notify_publication' in tt.objectIds():
        info['id'] = 'notify_publication'
        info['transition'] = 'publish'
        # send an email to the owner of an object when it is published
        info['expression'] = 'python:portal.portal_transitions.backgroundSendMail(object, \'notify_publication\', roles=[\'Owner\'])'
        tt.manage_addExpression(info)
    if not 'notify_rejection' in tt.objectIds():
        info['id'] = 'notify_rejection'
        info['transition'] = 'reject'
        # send an email to the owner of an object when it is rejected
        info['expression'] = 'python:portal.portal_transitions.backgroundSendMail(object, \'notify_rejection\', roles=[\'Owner\'])'
        tt.manage_addExpression(info)
    

def installSkins(self, out):
    # Go through the skin configurations and insert 'portal_transitions' into 
    # the configurations.  Preferably, this will right after where 'custom' is 
    # placed.  Otherwise, we append it to the end.
    skins_tool = getToolByName(self, 'portal_skins')
    # This is borrowed from CMFDefault/scripts/addImagesToSkinPaths.pys
    if 'portal_transitions' not in skins_tool.objectIds():
        # We need to add Filesystem Directory Views for any directories
        # in our skins/ directory.  These directories should already be
        # configured.
        addDirectoryViews(skins_tool, 'skins', GLOBALS)
        out.write("Added 'skins' directory view to portal_skins\n")
    skins = skins_tool.getSkinSelections()
    for skin in skins:
        path = skins_tool.getSkinPath(skin)
        path = [s.strip() for s in path.split(',') if s]
        new_skins = ['portal_transitions']

        new_skins.reverse()
        for n in new_skins:
            if n not in path:
                try: path.insert(path.index('custom')+1, n)
                except ValueError:
                    path.append(n)
                print >> out, "Added '%s' to %s skin\n" % (n,skin)
            else:
                print >> out, "Skipping '%s' for %s skin\n" % (n,skin)
        
        path = ', '.join(path)
        # addSkinSelection will replace existing skins as well.
        skins_tool.addSkinSelection(skin, path)


def uninstallSkins(self, out):
    # Go through the skin configurations and remove 'portal_transitions'
    skins_tool = getToolByName(self, 'portal_skins')
    skins = skins_tool.getSkinSelections()
    for skin in skins:
        path = skins_tool.getSkinPath(skin)
        path = [s.strip() for s in path.split(',') if s]
        new_skins = ['portal_transitions']
        for n in new_skins:
            if n in path:
                path.remove(n)
                print >> out, "Removing '%s' from %s skin\n" % (n,skin)
        path = ', '.join(path)
        # addSkinSelection will replace existing skins as well.
        skins_tool.addSkinSelection(skin, path)
    if 'portal_transitions' in skins_tool.objectIds():
        skins_tool.manage_delObjects(['portal_transitions'])


def patchWorkflowTool(self, out):
    # patch the workflow tool
    for name, newAttr in PatchedWorkflowTool.__dict__.items():
        # don't overwrite doc or module information
        if name not in ('__doc__', '__module__'):
            # save the old attribute as __monkey_name if exists
            # __dict__ doesn't show inherited attributes :/
            orig = getattr(WorkflowToolClass, name, None)
            if orig:
                stored_orig_name = '_monkey_%s' % name
                stored_orig = getattr(WorkflowToolClass, stored_orig_name, None)
                # don't double-patch!
                if stored_orig is None:
                    WorkflowToolClass.__dict__[stored_orig_name] = orig
                    print >> out, 'Patching workflow_tool.%s' % name
            # overwrite or add the new attribute
            WorkflowToolClass.__dict__[name] = newAttr


def unpatchWorkflowTool(self, out):
    # patch the workflow tool
    for name, newAttr in PatchedWorkflowTool.__dict__.items():
        # don't overwrite doc or module information
        if name not in ('__doc__', '__module__'):
            # save the old attribute as __monkey_name if exists
            # __dict__ doesn't show inherited attributes :/
            stored_orig_name = '_monkey_%s' % name
            stored_orig = getattr(WorkflowToolClass, stored_orig_name, None)
            if stored_orig:
                WorkflowToolClass.__dict__[name] = stored_orig
                del WorkflowToolClass.__dict__[stored_orig_name]
                print >> out, 'Restoring workflow_tool.%s' % name
            else:
                print >> out, 'Removing workflow_tool.%s' % name
                del WorkflowToolClass.__dict__[name]            
    
def install(self):
    out = StringIO()
    if not hasattr(self, 'portal_transitions'):
        addTool = self.manage_addProduct['PortalTransitions'].manage_addTool
        addTool('Portal Transitions Tool')
        out.write('Added Portal Transitions Tool\n')
    patchWorkflowTool(self, out)
    if INSTALL_EXAMPLE:
        installSkins(self, out)
        installTransitions(self, out)
    return out.getvalue()


def uninstall(self):
    out = StringIO()
    if hasattr(self, 'portal_transitions'):
        self.manage_delObjects(['portal_transitions'])
    unpatchWorkflowTool(self, out)
    if INSTALL_EXAMPLE:
        uninstallSkins(self, out)
    return out.getvalue()