import sys
from Acquisition import aq_base
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import CMFCorePermissions, Expression
from Products.CMFCore.WorkflowCore import ObjectDeleted, ObjectMoved, WorkflowException
from Products.PageTemplates.TALES import CompilerError
from Products.PortalTransitions.PortalTransitions import ANY_TYPE

class PatchedWorkflowTool:    
    def _executeExpressions(self, expressions, ob):
        portal = getToolByName(self, 'portal_url').getPortalObject()
        reindex = 1
        if expressions:
            for e in expressions:
                folder = ob.getParentNode()
                while not hasattr(aq_base(folder), 'isPrincipiaFolderish') and folder != portal:
                    folder = folder.getParentNode()
                expression_context = Expression.createExprContext(folder, portal, ob)
                try:
                    res = e._execute(expression_context)
                except ObjectDeleted, ex:
                    res = ex.getResult()
                    ob = None
                    reindex = 0
                except ObjectMoved, ex:
                    res = ex.getResult()
                    ob = ex.getNewObject()
                except:
                    getToolByName(self, 'plone_utils').logException()
            if reindex:
                self._reindexWorkflowVariables(ob)
        return ob
    
    
    # _invokeWithNotification is derived from the corresponding method in 
    # CMFCore/WorkflowTool.py.  This code is licensed under the Zope Public 
    # License (see LICENSE.ZPL)
    #security.declarePrivate( '_invokeWithNotification' )
    def _invokeWithNotification(self, wfs, ob, action, func, args, kw):
    
        """ Private utility method:  call 'func', and deal with exceptions
            indicating that the object has been deleted or moved.
        """
        reindex = 1
        for w in wfs:
            w.notifyBefore(ob, action)
        try:
            res = apply(func, args, kw)
            new_ob = ob
        except ObjectDeleted, ex:
            res = ex.getResult()
            new_ob = None
            reindex = 0
        except ObjectMoved, ex:
            res = ex.getResult()
            ob = ex.getNewObject()
            new_ob = ob
        except:
            exc = sys.exc_info()
            try:
                for w in wfs:
                    w.notifyException(ob, action, exc)
                raise exc[0], exc[1], exc[2]
            finally:
                exc = None
        for w in wfs:
            w.notifySuccess(ob, action, res)
        if reindex:
            self._reindexWorkflowVariables(ob)
        return (new_ob, res)
    
    
    # doActionFor is derived from the corresponding method in 
    # CMFCore/WorkflowTool.py.  This code is licensed under the Zope Public 
    # License (see LICENSE.ZPL)
    #security.declarePublic('doActionFor')
    def doActionFor(self, ob, action, wf_id=None, *args, **kw):
        """ Execute the given workflow action for the object.
        
        o Invoked by user interface code.
    
        o Allows the user to request a workflow action.
        
        o The workflow object must perform its own security checks.
        """

        # grab the workflow
        wfs = self.getWorkflowsFor(ob)
        if wfs is None:
            wfs = ()
        if wf_id is None:
            if not wfs:
                raise WorkflowException('No workflows found.')
            found = 0
            for wf in wfs:
                if wf.isActionSupported(ob, action):
                    found = 1
                    break
            if not found:
                raise WorkflowException(
                    'No workflow provides the "%s" action.' % action)
        else:
            wf = self.getWorkflowById(wf_id)
            if wf is None:
                raise WorkflowException(
                    'Requested workflow definition not found.')
    
        # We now have the workflow id and action in hand -- now get a list of
        # expressions to execute
        wf_id = wf.getId()
        types_tool = getToolByName(self, 'portal_types')
        type_id = getattr(types_tool.getTypeInfo(ob), 'id', None)
        expressions_before = []
        expressions_after = []
        try:
            tt = getToolByName(self, 'portal_transitions')
        except AttributeError, e:
            self.plone_log('Please install PortalTransitions using the QuickInstaller tool or remove the package from your Products directory and restart Zope.')
            raise
#        print 'type_id = %s, wf_id = %s, action = %s' % (type_id, wf_id, action)
        for e in tt.objectValues():
            if (e.type == ANY_TYPE or e.type is None or e.type == type_id) and \
               (e.workflow == wf_id) and \
               (e.transition == action):
                   if e.execute_after:
                       expressions_after.append(e)
                   else:
                       expressions_before.append(e)

        new_ob = self._executeExpressions(expressions_before, ob)
            
        (new_ob, result) = self._invokeWithNotification(
            wfs, new_ob, action, wf.doActionFor, (new_ob, action) + args, kw)

        #if result:
        #    result.reindexObjectSecurity()
    
        new_ob = self._executeExpressions(expressions_after, new_ob)
        if new_ob != ob:
            return new_ob
        else:
            return result
