import Globals
from AccessControl import ClassSecurityInfo
from OFS.OrderedFolder import OrderedFolder as BaseClass
from Products.CMFCore import CMFCorePermissions, Expression
from Products.CMFCore.utils import getToolByName, SimpleItemWithProperties
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.TALES import CompilerError
from Products.PortalTransitions import ANY_TYPE

from utils import delete_option



class TransitionExpression(SimpleItemWithProperties):
    """ """
    meta_type= 'Transition Expression'
    manage_options = SimpleItemWithProperties.manage_options

    security = ClassSecurityInfo()
    security.declareProtected(CMFCorePermissions.ManagePortal,
                              'manage_editProperties',
                              'manage_changeProperties',
                              'manage_propertiesForm',
                              )

    _properties = (
        {'id':'type', 'type':'string', 'mode':'w', 'label':'Type'},
        {'id':'workflow', 'type':'string', 'mode':'w', 'label':'Workflow'},
        {'id':'transition', 'type':'string', 'mode':'w', 'label':'Transition'},
        {'id':'expression', 'type':'string', 'mode':'w', 'label':'Expression'},
        {'id':'execute_after', 'type':'integer', 'mode':'w', 'label':'Execute after transition'},
        )

    type = ''
    workflow = ''
    transition = ''
    expression = ''
    _expr = None


    def __init__(self, id, **kw):
        self.id = id
        apply(self.manage_changeProperties, (), kw)


    def getTransitionCode(self):
        """Return the workflow_id, transition_id, and type_id as a single string code"""
        return self.getParentNode()._getTransitionCode(self.type, self.workflow, self.transition)


    def _updateProperty(self, id, value):
        if id == 'expression':
            self._expr = Expression.Expression(value)
        SimpleItemWithProperties._updateProperty(self, id, value)


    def _execute(self, expression_context):
        if not self._expr:
            raise ValueError, 'Empty expression for transition %s' % self.id
        self._expr(expression_context)


class PortalTransitions(BaseClass):
    """ """
    id = 'portal_transitions'
    title = 'Manage extra scripts executed at workflow transitions'
    meta_type= 'Portal Transitions Tool'

    typeClasses = [
        {'class':TransitionExpression,
         'name':TransitionExpression.meta_type,
         'action':'manage_addExpression',
         'permission':'Manage portal'},
        ]
    allowedTypes = [TransitionExpression.meta_type]

    security = ClassSecurityInfo()

    manage_options = ({'label':'Overview', 'action':'manage_overview' },) + \
                       BaseClass.manage_options
    manage_options = delete_option(manage_options, 'View')

    security.declareProtected( CMFCorePermissions.ManagePortal, 'manage_overview' )
    manage_overview = PageTemplateFile('www/manage_overview', globals())
    manage_overview.__name__ = 'manage_overview'

    security.declareProtected( CMFCorePermissions.ManagePortal, 'manage_main' )
    manage_main = PageTemplateFile('www/manage_main', globals())
    manage_main.__name__ = 'manage_main'

#    def __init__(self, id=None):
#        BaseClass.__init__(self, id)

    def all_meta_types(self):
        """Adds PortalTransitions-specific meta types."""
        all = PortalTransitions.inheritedAttribute('all_meta_types')(self)
        return tuple(self.typeClasses) + tuple(all)

    def filtered_meta_types(self, user=None):
        # Filters the list of available meta types.
        meta_types = []
        for meta_type in self.all_meta_types():
            if meta_type['name'] in self.allowedTypes:
                meta_types.append(meta_type)
        return meta_types


    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_addExpression')
    def manage_addExpression(self, info=None, REQUEST=None):
        """ Add a new transition expression"""
        if not info:
            info = self._getInfo(REQUEST, 'new')
        errors = {}
        err = self._validateExpression(info['expression'])
        if err:
            errors['new'] = err
        err = self._validateId(info['id'], check_for_collision=1)
        if err:
            errors['new'] = err
        if not errors.has_key('new'):
            ob = TransitionExpression(info['id'])
            self._setObject(ob.id, ob)
            ob.manage_changeProperties(**info)

        if REQUEST:
            return self.manage_main(REQUEST, errors=errors)
        else:
            if errors.has_key('new'):
                raise ValueError, errors['new']
            return ob

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_editExpressions')
    def manage_editExpressions(self, REQUEST):
        """Edit transition expressions"""
        errors = {}
        for id in self.objectIds():
            info = self._getInfo(REQUEST, id)
            err = self._validateExpression(info['expression'])
            if err:
                errors[id] = err
            err = self._validateId(info['id'], check_for_collision=0)
            if err:
                errors[id] = err
            if not errors.has_key(id):
                getattr(self, id).manage_changeProperties(**info)
                if not id == info['id']:
                    self.manage_renameObjects([id], [info['id']])
        return self.manage_main(REQUEST, errors=errors)

    def getTransitionCodes(self):
        """Return a list of available types+workflows+transitions for use in manage_main"""
        workflow_tool = getToolByName(self, 'portal_workflow')
        type_transitions = []
        
        for workflow in workflow_tool.objectValues():
            for transition_id in workflow.transitions.objectIds():
                type_transitions.append(self._getTransitionCode(None, workflow.getId(), transition_id))

        types_tool = getToolByName(self, 'portal_types')
        for type in types_tool.objectValues():
            workflows = workflow_tool.getWorkflowsFor(type.id)
            for workflow in workflows:
                for transition_id in workflow.transitions.objectIds():
                    type_transitions.append(self._getTransitionCode(type.getId(), workflow.getId(), transition_id))

        return type_transitions

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_move_objects_up')
    def manage_move_objects_up(self, REQUEST, ids=None, delta=1):
        """ Move specified sub-objects up by 1"""
        if ids is None:
            ids = self._getSelectedIds(REQUEST)
        return BaseClass.manage_move_objects_up(self, REQUEST, ids, delta)

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_move_objects_down')
    def manage_move_objects_down(self, REQUEST, ids=None, delta=1):
        """ Move specified sub-objects down by 1"""
        if ids is None:
            ids = self._getSelectedIds(REQUEST)
        return BaseClass.manage_move_objects_up(self, REQUEST, ids, delta)

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_move_objects_to_top')
    def manage_move_objects_to_top(self, REQUEST, ids=None):
        """ Move specified sub-objects to top of container."""
        if ids is None:
            ids = self._getSelectedIds(REQUEST)
        return BaseClass.manage_move_objects_to_top(self, REQUEST, ids)

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_move_objects_to_bottom')
    def manage_move_objects_to_bottom(self, REQUEST, ids=None):
        """ Move specified sub-objects to bottom of container."""
        if ids is None:
            ids = self._getSelectedIds(REQUEST)
        return BaseClass.manage_move_objects_to_bottom(self, REQUEST, ids)

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_delExpressions')
    def manage_delExpressions(self, REQUEST=None, ids=None):
        """ Delete selected expressions"""
        if ids is None:
            ids = self._getSelectedIds(REQUEST)
        self.manage_delObjects(ids)
        if REQUEST:
            REQUEST.RESPONSE.redirect('%s/manage_main' % self.absolute_url())

    def _getInfo(self, REQUEST, id):
        # Extract information about a particular transition expression from the REQUEST
        info = {}
        info['id'] = REQUEST.get('id_'+id)
        (info['type'], info['workflow'], info['transition']) = REQUEST['transition_code_'+id].split('/')
        if info['type'] == ANY_TYPE:
            info['type'] = None
        info['execute_after'] = REQUEST.get('execute_after_' + id, None)
        info['expression'] = REQUEST['expression_' + id].strip()
        return info

    def _validateId(self, id, check_for_collision):
        # utility method for validating ids
        if id == 'new':
            return 'The id \'new\' is reserved.  Please choose another.'
        # do basic id validation
        if not self.plone_utils.good_id(id):
            # id is bad
            return '\'%s\' is not a legal name.' % id
        # check for a catalog index
        if id in self.portal_catalog.indexes():
            return '\'%s\' is reserved.' % id
        if check_for_collision:
            if id in self.objectIds():
                return 'There is already an expression with id \'%s\'' % id


    def _validateExpression(self, expr):
        # utility method for validating expressions
        if not expr:
            return 'Please enter a TALES expression'
        try:
            Expression.Expression(expr)
        except CompilerError, e:
            return 'Please enter a valid expression (%s)' % str(e)

    def _getSelectedIds(self, REQUEST):
        # utility method that returns a list of selected items from the REQUEST
        ids = []
        for id in self.objectIds():
            if REQUEST.get('cb_' + id, None):
                ids.append(id)
        return ids

    def _getTransitionCode(self, type_id, workflow_id, transition_id):
        # utility method for creating a single string identifier that encodes 
        # type id, workflow id, and transition id
        if type_id is None:
            type_id = ANY_TYPE
        return type_id+'/'+workflow_id+'/'+transition_id
    
    
    def getMembers(self, ids=None, roles=None, object=None, REQUEST=None):
        """Get a list of members.  
        
           * If a list of member ids is passed in in ids, returns a list of the
             corresponding member objects.
           
           * If a list of roles is passed in, returns a list of members with
             the specified roles.
             
           * If a list of roles and an object are passed in, returns a list of
             members with the specified roles, either globally, or locally in
             the context of the object.
        """
        if REQUEST:
            raise Unauthorized, 'getMembers cannot be called via a URL'

        portal = getToolByName(self, 'portal_url').getPortalObject()
        membership_tool = getToolByName(portal, 'portal_membership')

        if ids is not None:
            members = [membership_tool.getMemberById(id) for id in ids]
            return [m for m in members if m]  # remove None

        if not roles:
            return []

        users_with_role = {}
        users = portal.acl_users.getUsers(__include_groups__=0)
        for u in users:
            ur = u.getRoles()
            for r in roles:
                if r in ur:
                    users_with_role[u.getUserName()] = 1

        if object:
            o = object
            while 1:
                if hasattr(o, 'users_with_local_role'):
                    for r in roles:
                        user_list = o.users_with_local_role(r)
                        for u in user_list:
                            user = portal.acl_users.getUserById(u, None)
                            if user:
                                if user.isGroup():
                                    for gu in user.getGroupUsers():
                                        users_with_role[gu.getUserName()] = 1
                                else:
                                    users_with_role[u] = 1
                if o == portal:
                    break
                o = o.getParentNode()
                

        members = []
        user_list = users_with_role.keys()
        user_list.sort()
        for u in user_list:
            member = membership_tool.getMemberById(u)
            if member is None:
                continue
            if object:  # make sure no PLIP 16 stuff eliminates roles locally
                ric = member.getRolesInContext(object)
                for r in roles:
                    if r in ric:
                        members.append(member)
                        break
            else:
                members.append(member)
        return members
    

    def sendMail(self, object, script_id, ids=None, roles=None, addresses=None,
                 REQUEST=None, **kwargs):
        if REQUEST:
            raise Unauthorized, 'getMembers cannot be called via a URL'
        o = None
        if roles:
            o = object
        recipients = self.getMembers(ids, roles, o)
        if addresses:
            recipients += list(addresses)
        mh = getToolByName(self, 'MailHost')
        
        script = getattr(object, script_id)
        for recipient in recipients:
            mh.send(script(recipient, **kwargs))


    def backgroundSendMail(self, object, script_id, ids=None, roles=None,
                           addresses=None, REQUEST=None, **kwargs):
        """Send emails via a background thread.  Useful if you have lots
        of things to send out."""
        if REQUEST:
            raise Unauthorized, 'getMembers cannot be called via a URL'
        tool_path = self.getPhysicalPath()
        object_path = object.getPhysicalPath()
        get_transaction().commit()  # commit to avoid read conflicts
        worker_thread = MailSender(tool_path, object_path, script_id, ids, roles,
                                   addresses, self.REQUEST, **kwargs)
        worker_thread.start()


import copy
import os
import sys
from threading import Thread
import Zope
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.BaseRequest import RequestContainer

class MailSender(Thread):
    
    def __init__(self, tool_path, object_path, script_id, ids, roles,
                 addresses, request, **kwargs):
        self.tool_path = tool_path
        self.object_path = object_path
        self.script_id = script_id
        self.ids = ids
        self.roles = roles
        self.addresses = addresses
        self.kwargs = kwargs
        self.environ = {}
        self.other = {}
        self.form = {}
        for k in request.environ.keys():
            if request.get(k) is not None:
                self.environ[k] = request.get(k)
        for k in request.other.keys():
            # only get canonical stuff -- avoid evil things added by
            # CMFFormController, PortalFactory, etc
            if k in ['VIRTUAL_URL_PARTS', 'AUTHENTICATED_USER', 'URL',
                     'VirtualRootPhysicalPath', 'VIRTUAL_URL', 'SERVER_URL',
                     'AUTHENTICATION_PATH', 'traverse_subpath',
                     'TraversalRequestNameStack'] \
                     or k.startswith('URL') or k.startswith('BASE'):
                if request.get(k) is not None:
                    self.other[k] = request.get(k)
        for k in request.form.keys():
            if request.get(k) is not None:
                self.form[k] = request.get(k)
        Thread.__init__(self)


    def run(self):
        app = Zope.app()
        resp = HTTPResponse(stdout=sys.stdout)
        env = copy.deepcopy(os.environ)
        env.update(self.environ)
        req = HTTPRequest(sys.stdin, env, resp)
        req.other.update(self.other)
        req.form.update(self.form)
        app = app.__of__(RequestContainer(REQUEST = req))

        tool = app.unrestrictedTraverse(self.tool_path)
        object = app.unrestrictedTraverse(self.object_path)
        tool.sendMail(object, self.script_id, self.ids, self.roles,
                      self.addresses, self.kwargs)
        app._p_jar.close()  # close the connection to the ZODB
        # make sure to clean up after ourselves
        tool = None
        object = None
        resp = None
        req = None
        env = None
        app = None


Globals.InitializeClass(PortalTransitions)
